package java.lang.ref;

import sun.misc.SharedSecrets;
import sun.misc.JavaLangRefAccess;
import sun.misc.Cleaner;

public abstract class Reference<T>
{
    private T referent;
    volatile ReferenceQueue<? super T> queue;
    volatile Reference next;
    private transient Reference<T> discovered;
    private static Reference.Lock lock;
    private static Reference<Object> pending;
    
    static boolean tryHandlePending(boolean b) {
        Reference<Object> pending;
        Cleaner cleaner;
        try {
            synchronized (Reference.lock) {
                if (Reference.pending != null) {
                    pending = Reference.pending;
                    cleaner = ((pending instanceof Cleaner) ? ((Cleaner)pending) : null);
                    Reference.pending = pending.discovered;
                    pending.discovered = null;
                }
                else {
                    if (b) {
                        Reference.lock.wait();
                    }
                    return b;
                }
            }
        }
        catch (OutOfMemoryError outOfMemoryError) {
            Thread.yield();
            return true;
        }
        catch (InterruptedException ex) {
            return true;
        }
        if (cleaner != null) {
            cleaner.clean();
            return true;
        }
        ReferenceQueue<Object> queue = pending.queue;
        if (queue != ReferenceQueue.NULL) {
            queue.enqueue(pending);
        }
        return true;
    }
    
    public T get() {
        return this.referent;
    }
    
    public void clear() {
        this.referent = null;
    }
    
    public boolean isEnqueued() {
        return this.queue == ReferenceQueue.ENQUEUED;
    }
    
    public boolean enqueue() {
        return this.queue.enqueue(this);
    }
    
    Reference(T t) {
        this(t, null);
    }
    
    Reference(T referent, ReferenceQueue<? super T> referenceQueue) {
        this.referent = referent;
        this.queue = ((referenceQueue == null) ? ReferenceQueue.NULL : referenceQueue);
    }
    
    static {
        Reference.lock = new Reference.Lock();
        Reference.pending = null;
        ThreadGroup threadGroup;
        for (threadGroup = (threadGroup = Thread.currentThread().getThreadGroup()); threadGroup != null; , threadGroup = threadGroup.getParent()) {}
        Reference.ReferenceHandler o = new Reference.ReferenceHandler(threadGroup, "Reference Handler");
        o.setPriority(10);
        o.setDaemon(true);
        o.start();
        SharedSecrets.setJavaLangRefAccess(new JavaLangRefAccess() {
            public boolean tryHandlePendingReference() {
                return Reference.tryHandlePending(false);
            }
        });
    }
    
    private static class Lock
    {
    }
    
    private static class ReferenceHandler extends Thread
    {
        private static void ensureClassInitialized(Class<?> clazz) {
            try {
                Class.forName(clazz.getName(), true, clazz.getClassLoader());
            }
            catch (ClassNotFoundException cause) {
                throw (Error)new NoClassDefFoundError(cause.getMessage()).initCause(cause);
            }
        }
        
        ReferenceHandler(ThreadGroup group, String name) {
            super(group, name);
        }
        
        @Override
        public void run() {
            while (true) {
                Reference.tryHandlePending(true);
            }
        }
        
        static {
            ensureClassInitialized(InterruptedException.class);
            ensureClassInitialized(Cleaner.class);
        }
    }
}
